@@ -6,11 +6,11 @@
 * Botan is released under the Simplified BSD License (see license.txt)
 */
 
-#ifndef BOTAN_P11_OBJECT_H__
-#define BOTAN_P11_OBJECT_H__
+#ifndef BOTAN_P11_OBJECT_H_
+#define BOTAN_P11_OBJECT_H_
 
 #include <botan/p11.h>
-#include <botan/p11_session.h>
+#include <botan/p11_types.h>
 #include <botan/secmem.h>
 
 #include <vector>
@@ -25,7 +25,7 @@ namespace PKCS11 {
 class Module;
 
 /// Helper class to build the Attribute / CK_ATTRIBUTE structures
-class BOTAN_DLL AttributeContainer
+class BOTAN_PUBLIC_API(2,0) AttributeContainer
    {
    public:
       AttributeContainer() = default;
@@ -33,14 +33,10 @@ class BOTAN_DLL AttributeContainer
       /// @param object_class the class type of this container
       AttributeContainer(ObjectClass object_class);
 
-      virtual ~AttributeContainer() = default;
+      virtual ~AttributeContainer() {}
 
-/* Microsoft Visual Studio <= 2013 does not support default generated move special member functions.
-   Everything else we target should support it */
-#if !defined( _MSC_VER ) || ( _MSC_VER >= 1900 )
       AttributeContainer(AttributeContainer&& other) = default;
       AttributeContainer& operator=(AttributeContainer&& other) = default;
-#endif
 
       // Warning when implementing copy/assignment: m_attributes contains pointers to the other members which must be updated after a copy
       AttributeContainer(const AttributeContainer& other) = delete;
@@ -128,7 +124,7 @@ class BOTAN_DLL AttributeContainer
    };
 
 /// Manages calls to C_FindObjects* functions (C_FindObjectsInit -> C_FindObjects -> C_FindObjectsFinal)
-class BOTAN_DLL ObjectFinder final
+class BOTAN_PUBLIC_API(2,0) ObjectFinder final
    {
    public:
       /**
@@ -139,17 +135,13 @@ class BOTAN_DLL ObjectFinder final
       ObjectFinder(Session& session, const std::vector<Attribute>& search_template);
 
       ObjectFinder(const ObjectFinder& other) = default;
-      ObjectFinder& operator=(const ObjectFinder& other) = default;
+      ObjectFinder& operator=(const ObjectFinder& other) = delete;
 
-/* Microsoft Visual Studio <= 2013 does not support default generated move special member functions.
-   Everything else we target should support it */
-#if !defined( _MSC_VER ) || ( _MSC_VER >= 1900 )
       ObjectFinder(ObjectFinder&& other) = default;
-      ObjectFinder& operator=(ObjectFinder&& other) = default;
-#endif
+      ObjectFinder& operator=(ObjectFinder&& other) = delete;
 
       /// Terminates a search for token and session objects (calls C_FindObjectsFinal)
-      ~ObjectFinder() BOTAN_NOEXCEPT;
+      ~ObjectFinder() noexcept;
 
       /**
       * Starts or continues a search for token and session objects that match a template, obtaining additional object handles (calls C_FindObjects)
@@ -173,7 +165,7 @@ class BOTAN_DLL ObjectFinder final
    };
 
 /// Common attributes of all objects
-class BOTAN_DLL ObjectProperties : public AttributeContainer
+class BOTAN_PUBLIC_API(2,0) ObjectProperties : public AttributeContainer
    {
    public:
       /// @param object_class the object class of the object
@@ -190,7 +182,7 @@ class BOTAN_DLL ObjectProperties : public AttributeContainer
    };
 
 /// Common attributes of all storage objects
-class BOTAN_DLL StorageObjectProperties : public ObjectProperties
+class BOTAN_PUBLIC_API(2,0) StorageObjectProperties : public ObjectProperties
    {
    public:
       /// @param object_class the CK_OBJECT_CLASS this storage object belongs to
@@ -237,7 +229,7 @@ class BOTAN_DLL StorageObjectProperties : public ObjectProperties
    };
 
 /// Common attributes of all data objects
-class BOTAN_DLL DataObjectProperties : public StorageObjectProperties
+class BOTAN_PUBLIC_API(2,0) DataObjectProperties final : public StorageObjectProperties
    {
    public:
       DataObjectProperties();
@@ -262,7 +254,7 @@ class BOTAN_DLL DataObjectProperties : public StorageObjectProperties
    };
 
 /// Common attributes of all certificate objects
-class BOTAN_DLL CertificateProperties : public StorageObjectProperties
+class BOTAN_PUBLIC_API(2,0) CertificateProperties : public StorageObjectProperties
    {
    public:
       /// @param cert_type type of certificate
@@ -318,7 +310,7 @@ class BOTAN_DLL CertificateProperties : public StorageObjectProperties
    };
 
 /// Common attributes of all key objects
-class BOTAN_DLL KeyProperties : public StorageObjectProperties
+class BOTAN_PUBLIC_API(2,0) KeyProperties : public StorageObjectProperties
    {
    public:
       /**
@@ -357,7 +349,7 @@ class BOTAN_DLL KeyProperties : public StorageObjectProperties
       */
       inline void set_allowed_mechanisms(const std::vector<MechanismType>&)
          {
-         throw Exception("Not implemented (KeyProperties::set_allowed_mechanisms)");
+         throw Not_Implemented("KeyProperties::set_allowed_mechanisms");
          }
 
       /// @return the key type of this key object
@@ -371,7 +363,7 @@ class BOTAN_DLL KeyProperties : public StorageObjectProperties
    };
 
 /// Common attributes of all public key objects
-class BOTAN_DLL PublicKeyProperties : public KeyProperties
+class BOTAN_PUBLIC_API(2,0) PublicKeyProperties : public KeyProperties
    {
    public:
       /// @param key_type type of key
@@ -424,10 +416,10 @@ class BOTAN_DLL PublicKeyProperties : public KeyProperties
       */
       inline void set_wrap_template(const AttributeContainer&)
          {
-         throw Exception("Not implemented (PublicKeyProperties::set_wrap_template)");
+         throw Not_Implemented("PublicKeyProperties::set_wrap_template");
          }
 
-      /// @param pubkey_info DER-encoding of the SubjectPublicKeyInfo for this public	key
+      /// @param pubkey_info DER-encoding of the SubjectPublicKeyInfo for this public key
       inline void set_public_key_info(const std::vector<uint8_t>& pubkey_info)
          {
          add_binary(AttributeType::PublicKeyInfo, pubkey_info);
@@ -435,7 +427,7 @@ class BOTAN_DLL PublicKeyProperties : public KeyProperties
    };
 
 /// Common attributes of all private keys
-class BOTAN_DLL PrivateKeyProperties : public KeyProperties
+class BOTAN_PUBLIC_API(2,0) PrivateKeyProperties : public KeyProperties
    {
    public:
       /// @param key_type type of key
@@ -503,10 +495,10 @@ class BOTAN_DLL PrivateKeyProperties : public KeyProperties
       */
       inline void set_unwrap_template(const AttributeContainer&)
          {
-         throw Exception("Not implemented (PrivateKeyProperties::set_unwrap_template)");
+         throw Not_Implemented("PrivateKeyProperties::set_unwrap_template");
          }
 
-      /// @param pubkey_info DER-encoding of the SubjectPublicKeyInfo for this public	key
+      /// @param pubkey_info DER-encoding of the SubjectPublicKeyInfo for this public key
       inline void set_public_key_info(const std::vector<uint8_t>& pubkey_info)
          {
          add_binary(AttributeType::PublicKeyInfo, pubkey_info);
@@ -514,7 +506,7 @@ class BOTAN_DLL PrivateKeyProperties : public KeyProperties
    };
 
 /// Common attributes of all secret (symmetric) keys
-class BOTAN_DLL SecretKeyProperties : public KeyProperties
+class BOTAN_PUBLIC_API(2,0) SecretKeyProperties final : public KeyProperties
    {
    public:
       /// @param key_type type of key
@@ -603,7 +595,7 @@ class BOTAN_DLL SecretKeyProperties : public KeyProperties
       */
       inline void set_wrap_template(const AttributeContainer&)
          {
-         throw Exception("Not implemented (SecretKeyProperties::set_wrap_template)");
+         throw Not_Implemented("SecretKeyProperties::set_wrap_template");
          }
 
       /**
@@ -614,12 +606,12 @@ class BOTAN_DLL SecretKeyProperties : public KeyProperties
       */
       inline void set_unwrap_template(const AttributeContainer&)
          {
-         throw Exception("Not implemented (SecretKeyProperties::set_unwrap_template)");
+         throw Not_Implemented("SecretKeyProperties::set_unwrap_template");
          }
    };
 
 /// Common attributes of domain parameter
-class BOTAN_DLL DomainParameterProperties : public StorageObjectProperties
+class BOTAN_PUBLIC_API(2,0) DomainParameterProperties final : public StorageObjectProperties
    {
    public:
       /// @param key_type type of key the domain parameters can be used to generate
@@ -638,7 +630,7 @@ class BOTAN_DLL DomainParameterProperties : public StorageObjectProperties
 /**
 * Represents a PKCS#11 object.
 */
-class BOTAN_DLL Object
+class BOTAN_PUBLIC_API(2,0) Object
    {
    public:
       /**
@@ -657,8 +649,8 @@ class BOTAN_DLL Object
       Object(Session& session, const ObjectProperties& obj_props);
 
       Object(const Object&) = default;
-      Object& operator=(const Object&) = default;
-      virtual ~Object() = default;
+      Object& operator=(const Object&) = delete;
+      virtual ~Object() {}
 
       /// Searches for all objects of the given type that match `search_template`
       template<typename T>
@@ -717,6 +709,14 @@ class BOTAN_DLL Object
          : m_session(session)
          {}
 
+      void reset_handle(ObjectHandle handle)
+         {
+         if(m_handle != CK_INVALID_HANDLE)
+            throw Invalid_Argument("Cannot reset handle on already valid PKCS11 object");
+         m_handle = handle;
+         }
+
+   private:
       const std::reference_wrapper<Session> m_session;
       ObjectHandle m_handle = CK_INVALID_HANDLE;
    };
